// WattMètre/Chargeur pour BBS02 avec Lcd, RTC3231+EEPROM+ADS1115
// Auteur : Jean-Noël MICHEL de la Rochefoucauld
// Date   : 09/08/2018
// Version: V 1.0 avec Inter MosFet
//        : V 2.0 avec Relais au lieu du MosFet qui fait des retours de courant !

#include <avr/wdt.h>                    // WatchDog
#include <Wire.h>                       // I2C standard
#include <EEPROM.h>                     // Eeprom
#include <LiquidCrystal_I2C.h>          // Afficheur LCD sur I2C
#include "ds3231.h"                     // Horloge DS3231

#define ulong uint32_t

LiquidCrystal_I2C lcd(0x3F,16,2);       // i2c adresse du LCD en 0x3F en 16C x 2L
const byte   Inter   = 2;               // Relais Chargeur Sortie 2
const byte   Buzzer  = 6;               // Buzzer Son      Sortie 6

// Variable Système
bool   Aff1ou2 = false;                 // Affichage de l'écran 1 ou 2 sur le LCD
bool   Charge  = false;                 // Charge activé ou non
bool   Pleine  = false;                 // Batterie Pleine, plus de Charge
byte   i, j;                            // Variable de Boucle ou général
byte   NumInd  = 0;                     // Numéro de la mesure par 100ms (0 à 9)
word   pos;                             // Position dans l'EEPROM
word   CourRef;                         // Valeur de la mesure à vide mesure Courant (2,5v)
ulong  Tempo1;                          // Pour le temps des 100 millisecondes
ulong  Tempo1s;                         // Pour le temps en seconde
ulong  Tempo1m;                         // Pour le temps en minute
struct ts t;                            // Pour l'Horloge
String Message;                         // Message à Afficher

// Variable Sauvé EEPROM
byte   NumListe;                        // Numéro dans la Liste (0 à 99)
word   NumCharge;                       // Numéro de la Charge (0 à 65535)
word   NumEnr;                          // Numéro d'enregistrement EEPROM Externe (0 à 32766)
word   NumPos;                          // Numéro de Position dans EEPROM Externe (0 à 32766)
ulong  DureeTot;                        // Durée Total de Charge de la Batterie (0000h)
float  WattTotal;                       // Watt Total Chargé (00000,0W)
float  AhTotal;                         // Total Chargé en Ampère (00000,0Ah)
float  SeuilCharge;                     // Seuil de Charge, Coupe la Charge (protection Batterie)

// Variable Programme
byte   Second, Minute, Heure;           // Du RTS3231
byte   Semaine, Jour, Mois;             // Du RTS3231
int    Annee;                           // Du RTS3231
word   Duree     = 0;                   // Durée de Charge en Second
float  Courant   = 0.0;                 // Courant   (0,000)
float  CourOld   = 0.0;                 // Ancien Courant pour Delta / Min
float  CourMax   = 0.0;                 // Courant Maxi (0,000)
float  TensionB  = 0.0;                 // Tension Batterie  (00,00)
float  TensionC  = 0.0;                 // Tension Chargeur  (00,00)
float  TensOld   = 54.5;                // Ancienne Tension pour Delta / Min
float  TensMax   = 0.0;                 // Tension Maxi (00,00)
float  Tension5v = 0;                   // Tension du 5V de l'Arduino
float  AhTot     = 0;                   // Total Chargé en Ampère
float  WattTot   = 0;                   // Watt Total Chargé (000,0)
float  TabTens[12];                     // Tableau des Mesures de Tension
float  TabCour[12];                     // Tableau des Mesures de Courant

#include "Peripherique.h"               // Périphérique du Programme ADS1115, RTS3231, 24C256
#include "AfficheLcd.h"                 // Affichage LCD 2 ligne 16 Car
#include "Mesure.h"                     // Mesure du WattMètre sur ADS1115

// Initilisation
void setup()
{
  // Initialisation du WatchDog. Se déclenche au bout de 8 secondes d'inactivité
  wdt_enable(WDTO_8S);

  // Initialisation du port I2C
  Wire.begin();
  DS3231_init(4);                       // Initialisation de l'Horloge DS3231

  // Initialisation la liaison Série (Lecture EEPROM)
  Serial.begin(115200);                 // Initialise Serial à 115200 baud (retour PC)

  // Init Buzzer
  pinMode(Buzzer, OUTPUT);              // Définir la sortie du Buzzer
  digitalWrite(Buzzer, LOW);            // Silence

  // Init Relais du Chargeur
  pinMode(Inter, OUTPUT);               // Définir la sortie du Relais
  digitalWrite(Inter, HIGH);            // Coupé

  // Initialisation de l'Écran LCD
  lcd.init();
  delay(100);
  lcd.backlight();
  lcd.setCursor(0,0);
  lcd.print("WattMetre JNM V2");

  // Mesure de la Tension du Chargeur
  MesureTensionC();                     // Mesure de la Tension au démarrage
  lcd.setCursor(0,1);                   // Passe à la deuxième ligne (1)
  lcd.print("Chargeur  ");
  lcd.print(String(TensionC,2));
  lcd.print("V");
  wdt_reset();                          // Reset du watchDog : C'est reparti pour 8 secondes

  // Son pour attirer la lecture de l'Affichage
  tone(Buzzer, 440, 100);
  delay(100);
  tone(Buzzer, 880, 100);
  delay(100);
  tone(Buzzer, 440, 100);
  delay(4000);                          // Le temps de lire les infos

  // Détermine le Zéro de l'AmpèreMètre (2,5v)
  CourRef = lireADC(1, 1);              // Première mesure de Référence (2,5v)
  word val;
  for (i = 0; i < 30; ++i)
  {
    val = lireADC(1, 1);                // Lecture à vide de l'AmpèreMètre
    if (val < CourRef) CourRef = val;   // Cherche la plus petite valeur
    delay(50);
  }

  // Remise à Zéro des totaux
  WattTot = 0;
  AhTot   = 0;
  Duree   = 0;
  TensMax = 0;
  for (i = 0; i < 12; i++) { TabTens[i] = TensionB; TabCour[i] = 0; }

  // Lire les Variables Sauvées sur l'EEPROM Interne
  LectureEepromI();
  /// Initialisation des donnée dans l'EEPROM (à commenter après l'enregistrement)
/*
  NumCharge = 0;                        // Numéro de la Charge (stoker en EEPROM)
  WattTotal = 2295;                     // Watt Total Chargé (000000,0)
  AhTotal   = 32.5;                     // Total Chargé en Ampère (une charge complète + 1/2)
  DureeTot  = 64800;                    // Durée Total de Charge de la Batterie (18h)
  SeuilCharge = 53.8;                   // Seuil de Coupure de la Charge
*/  /// A commenter après première init

  // Afficher les données Sauver en EEPROM Interne
  lcd.setCursor(0,0);                   // Première Ligne (0)
  lcd.print("Nc:");
  lcd.print(String(NumCharge + 1));
  lcd.print(" S:");
  lcd.print(String(SeuilCharge,2));
  lcd.print("V   ");
  lcd.setCursor(0,1);                   // Passe à la deuxième ligne (1)
  lcd.print(String(WattTotal, 0));
  lcd.print("Wt Tps:");
  lcd.print(String(DureeTot / 3600));
  lcd.print("h  ");

  wdt_reset();                          // Reset du watchDog : C'est reparti pour 8 secondes

  // Son pour attirer la lecture de l'Affichage
  tone(Buzzer, 440, 100);
  delay(100);
  tone(Buzzer, 880, 100);
  delay(100);
  tone(Buzzer, 440, 100);
  delay(5000);                          // Boucle d'attente de 5 seconds

  // Début la boucle des temps
  Tempo1  = millis();
  Tempo1s = millis();
  Tempo1m = millis();
}


/// -------------------
/// Programme Principal
/// -------------------
void loop()
{
  wdt_reset();                          // Reset du watchDog : C'est reparti pour 8 secondes

  // -----------------------------------------
  // Action à faire toute les 100 milliseconds
  // -----------------------------------------
  if ((millis() - Tempo1) > 100)        // Gestion des Mesures
  {
    Tempo1 = millis();
    MesureTensionB();                   // Mesure de la tension Batterie
    MesureTensionC();                   // Mesure de la tension du Chargeur
    MesureCourant();                    // Mesure du Courant de Charge
    TabTens[NumInd] = TensionB;         // Stocke la Valeur de Tension Batterie
    TabCour[NumInd] = Courant;          // Stocke la Valeur de Courant
    if (Courant  > CourMax) CourMax = Courant;  // Stocke le Courant Maxi
    if (TensionB > TensMax) TensMax = TensionB; // Stocke la Tension Maxi
    ++NumInd;
  }

  // ----------------------------------
  // Action à faire toute les 1 Seconds
  // ----------------------------------
  if ((millis() - Tempo1s) > 1000)      // Gestion des Mesures par Seconds
  {
    Tempo1s = millis();                 // Mémorisation pour le prochaine calcul
    DateHeure();                        // Lecture de la Date et Heure

    // Calcul la moyenne des 10 mesures faite toute les 100ms
    float TotT = 0, TotC = 0;
    for (i = 0; i < NumInd; i++)
    {
      TotT += TabTens[i];
      TotC += TabCour[i];
    }
    TensionB = TotT / NumInd;
    Courant  = TotC / NumInd;
    NumInd = 0;                         // Remise à Zéro du Compteur de 100 ème

    // Cumul des Valeurs
    if (Charge)
    {
      AhTot   += Courant / 3600;
      WattTot += TensionB * (Courant / 3600);

      // Incrémente le Temps de Charge
      ++Duree;                          // Augmente la Durée de Charge si Charge +1s

      // Affichage sur l'écran
      if (Aff1ou2) AfficherT();
      else         AfficherM();
      Aff1ou2 = ! Aff1ou2;                // On inverse l'affichage
    }
    else
    AfficheNoCharge();
  }

  // ---------------------------------
  // Action à faire après 10s du début
  // ---------------------------------
  if (! Charge & ! Pleine & (millis() - Tempo1m > 10000))
  {
    // Détection de Début de Charge si Chargeur branché (pas l'Arduino seul)
    if (TensionC > 30.0)
    {
      digitalWrite(Inter, LOW);         // Activé l'interupteur : Charge en cours
      delay(1000);                      // Attendre Stabilisation 1 seconde
      MesureTensionB();                 // Mesure de la Tension Batterie
      MesureTensionC();                 // Mesure de la Tension Chargeur
      if (TensionB < TensionC)          // Début de Charge ?
      {
        Charge = true;                  // La batterie est en cours de Charge
        ++NumCharge;                    // Incrément du compteur de Charge
        NumEnr += (NumPos * 4) + 2;     // Détermine le nouvelle emplacement des Données
        if ((NumEnr + 2400) > 32766)    // Fin de l'EEPROM Externe (Si plus de 10h de Charge)
        {
          NumEnr = 0;                   // On écrase les anciennes Données
          NumPos = 0;
        }
        ++NumListe;                     // Incrément du compteur de Liste
        if (NumListe > 99) NumListe = 0;// Si fin de la liste écraser les premiers

        SauveCharge();                  // Init le journal des Charges en EEPROM

        // Son pour indiquer le début de Charge
        tone(Buzzer, 440, 100);
        delay(100);
        tone(Buzzer, 660, 100);
        delay(100);
        tone(Buzzer, 880, 100);
        delay(100);
        tone(Buzzer, 1010, 100);
        delay(100);
        tone(Buzzer, 1230, 100);
      }
    }
    else if (Serial.available() && ! Pleine) // Arduino seul, alors on décharge les EEPROM's
    {
      AfficheTrans();
      // Son pour indiquer le début du Transfert vers le PC
      tone(Buzzer, 440, 100);
      delay(100);
      tone(Buzzer, 660, 100);
      delay(100);
      tone(Buzzer, 880, 100);
      delay(100);
      tone(Buzzer, 660, 100);
      delay(100);
      tone(Buzzer, 440, 100);
      Pleine = true;
      Serial.print("Nombre de Charge : ");
      Serial.println(NumListe);
      for (word n = 0; n < NumListe; n++)
      {
        // Récupère les données dans l'EEPROM
        word pos = (n * 20) + 25;       // n de 0 à 99 * 20 = 2000 valeurs
        word date, heur;                // Date et Heure au format compressé
        EEPROM.get(pos     , date);     // Date
        EEPROM.get(pos +  2, heur);     // Heure
        EEPROM.get(pos +  4, Duree);    // Temps de Charge en Seconds
        EEPROM.get(pos +  6, TensMax);  // Tension Maxi
        EEPROM.get(pos + 10, AhTot);    // Cumul Ah
        EEPROM.get(pos + 14, WattTot);  // Cumul Watt
        EEPROM.get(pos + 18, NumEnr);   // Num Enr du détail (00000 à 32767) EEPROM Externe
        Annee   = date / 375;           // Calcul la Date
        Mois    = (date - (Annee * 375)) / 31;
        Jour    = (date - (Annee * 375)) % 31;

        // Envoie les informations d'Entête sur le Port Série
        Serial.println();
        Serial.println();
        Message = "Charge numéro : " + String(n);
        Serial.println(Message);
        Message = "Date du       : " + Ajout0(Jour) + "/" + Ajout0(Mois) + "/" + String(Annee) + " début à ";
        Heure   =   heur / 1800;        // Calcul l'Heure
        Minute  =  (heur - (Heure * 1800)) / 60;
        Second  = ((heur - (Heure * 1800)) % 30) * 2;
        Message += Ajout0(Heure) + ":" + Ajout0(Minute) + ":" + Ajout0(Second);
        Serial.println(Message);
        Message = "Durée         : " + String(Duree) + " seconds";
        Serial.println(Message);
        Message = "Tesnsion Max  : " + String(TensMax) + " V";
        Serial.println(Message);
        Message = "Ah Total      : " + String(AhTot) + " Ah";
        Serial.println(Message);
        Message = "Watt Max      : " + String(WattTot) + " W";
        Serial.println(Message);
        Serial.println();

        // Calcul les Tensions et Courants
        int DeltaT, DeltaC;
        Serial.print("Nombre d'enregistrement : ");
        Serial.println(NumEnr);
        for (word k = 0; k < NumEnr; k++)
        {
          wdt_reset();                  // Reset du watchDog : C'est reparti pour 8 secondes
          // Lire dans l'EEPROM les Données de Charge toute les Minutes
          pos = (k * 4) + NumEnr;
          // Lire les donnée dans l'EEPROM
          DeltaT = readEEPROM(pos);     // Delta de Tension
          DeltaC = readEEPROM(pos + 2); // Delta de Courant
          Message  = String(k) + ";";   // Numéro d'enregistrement
          Message += String(DeltaT / 1000.0, 3) + ";";// Soit pour 1,125V -> 1125V
          Message += String(DeltaC / 1000.0, 3);      // Soit pour 1,125A -> 1125A
          Serial.println(Message);
        }
        Serial.println("Fin de liste.");
      }
      // Son pour indiquer la fin du Transfert vers le PC
      tone(Buzzer, 440, 100);
      delay(100);
      tone(Buzzer, 660, 100);
      delay(100);
      tone(Buzzer, 888, 100);
      delay(100);
      tone(Buzzer, 660, 100);
      delay(100);
      tone(Buzzer, 440, 100);
      delay(2000);
    }
  }

  // --------------------------------
  // Action à faire toute les Minutes
  // --------------------------------
  if (millis() - Tempo1m > 60000)
  {
    Tempo1m = millis();                 // Sauve le temps

    // Vérifie que la Charge Arrive à la Fin et Sauve les infos
    if (Charge & ! Pleine & (TensionB > SeuilCharge))
    {
      digitalWrite(Inter, HIGH);        // Coupe la Charge, car Batterie Pleine
      Charge = false;                   // Indique que l'inter est coupé
      Pleine = true;                    // Indique le fin de Charge

      // Mettre à jour les Grands Totaux
      WattTotal += WattTot;
      AhTotal   += AhTot;
      DureeTot  += Duree;

      // Sauver les données en EEPROM Interne
      SauveCharge();                    // Les données de la Charge
      SauveEepromI();                   // Les Données Globales

      // Son pour indiquer la fin de charge
      tone(Buzzer, 1230, 100);
      delay(100);
      tone(Buzzer, 1010, 100);
      delay(100);
      tone(Buzzer, 888, 100);
      delay(100);
      tone(Buzzer, 660, 100);
      delay(100);
      tone(Buzzer, 440, 100);
    }

    // Si la charge est en cours, on sauve les infos dans l'EEPROM Externe
    if (Charge)
    {
      // Sauve dans l'EEPROM les Données de Charge toute les Minutes
      pos = (NumPos * 4) + NumEnr;
      int DeltaT, DeltaC;
      DeltaT = float((TensionB - TensOld) * 1000); // Soit pour 1,125V 1125V afin d'avoir la précision
      DeltaC = float((Courant  - CourOld) * 1000); // Soit pour 1,125A 1125A afin d'avoir la précision
      TensOld = TensionB; CourOld = Courant;// Pour Calculer les Deltas suivant

      // Sauver les donnée dans l'EEPROM
      writeEEPROM(pos    , DeltaT);
      writeEEPROM(pos + 2, DeltaC);
      ++NumPos;
      if ((NumPos * 4) + NumEnr > 32766)// Fin de l'EEPROM (32768)
      {
        NumEnr = 0;                     // On écrase les première Données
        NumPos = 0;
        SauveCharge();                  // Les données de la Charge
        SauveEepromI();                 // Les Données Globales
      }

      // Un petit son toute les Minutes
      tone(Buzzer, 440, 100);
    }
  }
  // Fin de la Boucle des temps
}
