// Sous-programme du WattMètre - Périphérique : ADC, Horloge, EEPRO.

/// -----------------------------------------------------------
/// ADS1115 Convertisseur Analogique Numérique 16 bits 4 Canaux
/// -----------------------------------------------------------
// Lire un Canal sur l'ADS1115
word lireADC(byte gain, byte canal)
{
  if (canal > 3) return 0;              // Maxi 4 Canaux (0 à 3)
  word configR = 0x0003 |               // Désactive le comparateur (défaut)
             //  0x0000 |               // Pas de mémorisation (défaut)
             //  0x0000 |               // Alerte Prêt Activé en bas (défaut)
             //  0x0000 |               // Traditionnel Comparateur (défaut)
                 0x0080 |               // 1600 mesures par second (défaut)
                 0x0100;                // Mode Simple Mesure (défaut)

  // Sélection du Gain
  switch (gain)
  {
    case (0):             // 0x0000 +/- 6.144V soit 1 bit = 0.1875mV    = Gain 0 (réel 2/3)
      break;
    case (1):             // 0x0200 +/- 4.096V soit 1 bit = 0.125mV     = Gain 1
      configR |= 0x0200;
      break;
    case (2):             // 0x0400 +/- 2.048V soit 1 bit = 0.0625mV    = Gain 2 (défaut)
      configR |= 0x0400;
      break;
    case (4):             // 0x0600 +/- 1.024V soit 1 bit = 0.03125mV   = Gain 4
      configR |= 0x0600;
      break;
    case (8):             // 0x0800 +/- 0.512V soit 1 bit = 0.015625mV  = Gain 8
      configR |= 0x0800;
      break;
    case (16):            // 0x0A00 +/- 0.256V soit 1 bit = 0.0078125mV = Gain 16
      configR |= 0x0A00;
  }

  // Sélection du Canal
  switch (canal)
  {
    case (0):
      configR |= 0x4000;
      break;
    case (1):
      configR |= 0x5000;
      break;
    case (2):
      configR |= 0x6000;
      break;
    case (3):
      configR |= 0x7000;
  }
  configR |= 0x8000;                     // Simple conversion

  // Écrire la config dans l'ADC et déclenche la Conversion
  Wire.beginTransmission(0x48);         // Active l'ADS1115
  Wire.write(0x01);                     // Écrire sur le Registre de l'ADS1115
  Wire.write((byte)(configR >> 8));     // MSB du Registre
  Wire.write((byte)(configR & 0xFF));   // LSB du Registre
  Wire.endTransmission();
  delay(8);                             // Attente de la conversion soit complète 8 ms

  // Lire une valeur (word) sur l'ADS1115
  Wire.beginTransmission(0x48);         // Active l'ADS1115
  Wire.write(0x00);                     // Active la Mesure
  Wire.endTransmission();
  Wire.requestFrom(0x48, 2);            // Demande de Résultat sur 2 octets
  return ((Wire.read() << 8) | Wire.read()); // Lire le Résultat (16 bits)
}


/// ----------------------
/// Horloge DS3231 sur I2C
/// ----------------------
void DateHeure()
{
  DS3231_get(&t);                       // Lire la Date - Heure
  Second  = t.sec;
  Minute  = t.min;
  Heure   = t.hour;
  Jour    = t.mday;
  Mois    = t.mon;
  Annee   = t.year;
  Semaine = t.wday;
}


/// ---------------------
/// EEPROM 24C256 sur I2C
/// ---------------------
// Écriture sur l'EEPROM d'un Entier
void writeEEPROM(word eadr, int data)
{
  i = data >> 8;
  j = data & 0xFF;
  Wire.beginTransmission(0x50);         // Demande d'Écriture EEPROM Externe
  Wire.write((word)(eadr >> 8));        // MSB de l'adresse
  Wire.write(eadr & 0xFF);              // LSB de l'adresse
  Wire.write(i);                        // Écriture de la donnée Haute
  Wire.endTransmission();
  delay(5);
  eadr++;
  Wire.beginTransmission(0x50);         // Demande d'Écriture EEPROM Externe
  Wire.write((word)(eadr >> 8));        // MSB de l'adresse
  Wire.write(eadr & 0xFF);              // LSB de l'adresse
  Wire.write(j);                        // Écriture de la donnée Basse
  Wire.endTransmission();
  delay(5);
}

// Lecture sur l'EEPROM d'un Entier
int readEEPROM(word eadr)
{
  Wire.beginTransmission(0x50);         // Demande le lecture EEPROM Externe
  Wire.write((word)(eadr >> 8));        // MSB de l'adresse
  Wire.write(eadr & 0xFF);              // LSB de l'adresse
  Wire.endTransmission();
  Wire.requestFrom(0x50, 2);            // Demande de lecture d'un Entier
  if (Wire.available()) i = Wire.read();// Lecture de la donnée (i)
  if (Wire.available()) j = Wire.read();// Lecture de la donnée (j)
  return (i << 8 | j);                  // Envoie le résultat
}


/// ---------------------------------
/// Gestion des Données dans l'EEPROM
/// ---------------------------------
// Sauve la Charge sur l'EEPROM Interne Position de 20 à 2020 sur 2048 (2000 valeurs)
void SauveCharge()
{
  // 20 octets : date, heure, Durée, TensionMax, AhTot, WattTot, NumEnr
  //             0/1   2/3    4/5    6-9         10-13  14-17    18/19
  DateHeure();                          // Lire la Date - Heure
  word date = ((Annee - 2000) * 375) + (Mois * 31) + Jour;   // la Date
  word heur = (Heure * 1800) + (Minute * 30) + (Second / 2); // l'Heure
  pos     = (NumListe * 20) + 25;       // 0 à 99 * 20 = 2000 valeurs
  EEPROM.put(pos     , date);           // Date
  EEPROM.put(pos +  2, heur);           // Heure
  EEPROM.put(pos +  4, Duree);          // Temps de Charge en Seconds
  EEPROM.put(pos +  6, TensMax);        // Tension Maxi
  EEPROM.put(pos + 10, AhTot);          // Cumul Ah
  EEPROM.put(pos + 14, WattTot);        // Cumul Watt
  EEPROM.put(pos + 18, NumEnr);         // Num Enr du détail (00000 à 32767) EEPROM Externe
}

// Sauve les Variables Globales de Stockage en EEPROM Interne (23 octets)
void SauveEepromI()
{
  EEPROM.put( 0, NumListe);             // Numéro de Liste                      (1 octets)
  EEPROM.put( 1, NumCharge);            // Numéro de la Charge                  (2 octets)
  EEPROM.put( 3, WattTotal);            // Watt Total Depuis Début              (4 octets)
  EEPROM.put( 7, AhTotal);              // Total Chargé en Ampère               (4 octets)
  EEPROM.put(11, DureeTot);             // Durée Total de Charge de la Batterie (4 octets)
  EEPROM.put(15, NumEnr);               // Numéro d'Enregistrement EEPROM Ext   (2 octets)
  EEPROM.put(17, NumPos);               // Nombre d'entrée du Détail (à x 4)    (2 octets)
  EEPROM.put(19, SeuilCharge);          // Seuil de charge (53,8V)              (4 octets)
}

// Lecture des valeurs sauvegardés en EEPROM Interne (23 octets)
void LectureEepromI()
{
  EEPROM.get( 0, NumListe);             // Numéro de Liste                      (1 octets)
  EEPROM.get( 1, NumCharge);            // Numéro de Charge                     (2 octets)
  EEPROM.get( 3, WattTotal);            // Watt Total depuis le début           (4 octets)
  EEPROM.get( 7, AhTotal);              // Ampère par Heure depuis le début     (4 octets)
  EEPROM.get(11, DureeTot);             // Temps de Charge en heure             (4 octets)
  EEPROM.get(15, NumEnr);               // Numéro de l'emplacement du Détail    (2 octets)
  EEPROM.get(17, NumPos);               // Nombre d'entrée du Détail (à x 4)    (2 octets)
  EEPROM.get(19, SeuilCharge);          // Seuil de charge (53,8V)              (4 octets)
}
